from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Literal, Optional, Union

from pydantic import BaseModel

from openreward.api.rollouts.serializers.models import NormalizedType


class RolloutConfig(BaseModel):
    run_name: str
    rollout_name: Optional[str] = None
    environment: Optional[str] = None
    split: Optional[str] = None
    metadata: Optional[dict] = None
    task_spec: Optional[Dict[str, Any]] = None


@dataclass
class Config:
    process_name: str
    shutdown_timeout: float
    send_loop_config: "SendLoopConfig"


@dataclass
class SendLoopConfig:
    max_items: int # max item threshold
    max_bytes: int # max byte threshold
    max_age: float # max age
    jitter: float # % jitter of the flush interval

    ring_capacity: int # max items, at which the ring buffer will be flushed

    max_batch_items: int # max number of items to flush at once
    max_batch_bytes: int # max number of bytes to flush at once

    max_retries: int # max number of retries
    backoff_base: float # base time of the backoff
    backoff_factor: float # factor of the backoff
    backoff_cap: float # cap time of the backoff

    max_upload_concurrency: int

    api_key: str | None
    base_url: str

@dataclass
class LogMessageEvent:
    # rollout info
    eventId: str
    timestamp: int
    index: int

    rolloutEventId: str

    # normalized event info
    type: NormalizedType
    content: Optional[str] = None # visible text or JSON string for tools
    contentReference: Optional[str] = None # only for hidden reasoning
    summary: Optional[str] = None # reasoning-only
    name: Optional[str] = None # tool name (tool_call only)
    callId: Optional[str] = None # join key for tool_call/result

    # extra info
    reward: Optional[float] = None
    is_finished: Optional[bool] = None
    metadata: Optional[Dict[str, Any]] = None

    eventType: Literal["message"] = "message"


@dataclass # Do we want this to be on the background process?
class RolloutStartedEvent:
    # rollout info
    eventId: str
    timestamp: int

    # rollout info
    runName: str
    step: Optional[int] = None
    environment: Optional[str] = None
    rolloutName: Optional[str] = None
    split: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    task_spec: Optional[Dict[str, Any]] = None

    eventType: Literal["rollout"] = "rollout"

    @classmethod
    def from_config(cls, event_id: str, timestamp: int, config: RolloutConfig, step: Optional[int] = None):
        return cls(
            eventId=event_id,
            timestamp=timestamp,
            runName=config.run_name,
            rolloutName=config.rollout_name,
            environment=config.environment,
            split=config.split,
            metadata=config.metadata,
            task_spec=config.task_spec,
            step=step
        )

@dataclass
class FlushEvent:
    event_type: Literal["flush"] = "flush"

@dataclass
class ShutdownEvent:
    event_type: Literal["shutdown"] = "shutdown"

InputEvent = Union[LogMessageEvent, RolloutStartedEvent, FlushEvent, ShutdownEvent]

@dataclass
class ShutdownResponse:
    success: bool
    timestamp: datetime
    event_type: Literal["shutdown_response"] = "shutdown_response"

OutputEvent = ShutdownResponse


import os
from typing import Optional
from urllib.parse import urlparse, urlunparse


from openreward.models import Config, SendLoopConfig
from openreward.api.rollouts.rollout import RolloutAPI
from openreward.api.environments.client import EnvironmentsAPI
from openreward.api.sandboxes import SandboxesAPI, SandboxSettings

OPENREWARD_API_KEY_ENV_VAR_NAME = "OPENREWARD_API_KEY"
DEFAULT_BASE_URL = "https://openreward.ai"

def _prepend_subdomain(url: str, subdomain: str) -> str:
    parsed = urlparse(url)
    new_netloc = f"{subdomain}.{parsed.netloc}"
    return urlunparse(parsed._replace(netloc=new_netloc))

class OpenReward:
    def __init__(self, api_key: str | None = None, base_url: str | None = None):
       
        self.api_key = api_key or os.getenv(OPENREWARD_API_KEY_ENV_VAR_NAME, "")
        assert self.api_key is not None

        base_url = base_url or os.getenv("OPENREWARD_URL", DEFAULT_BASE_URL)
        assert base_url is not None
        
        self.base_url = _prepend_subdomain(base_url, "api")
        self.environments_base_url = _prepend_subdomain(base_url, "matrix")
        self.sandboxes_base_url = _prepend_subdomain(base_url, "construct")

        self.config = Config(
            process_name="openreward_client", 
            shutdown_timeout=10.0, 
            send_loop_config=SendLoopConfig(
                max_items=128,
                max_bytes=4_000_000, # 4MB
                max_age=1.0, # 1 second
                jitter=0.05, # 5% jitter
                ring_capacity=100000, # 10,000 items
                max_batch_items=1024,
                max_batch_bytes=8_000_000, 
                max_retries=4, 
                backoff_base=0.5, # 0.5 seconds
                backoff_factor=2.0, 
                backoff_cap=30.0, # 30 seconds
                max_upload_concurrency=1024,
                api_key=self.api_key,
                base_url=self.base_url
            )
        )
        
        self._rollout_api: Optional[RolloutAPI] = None
        self._environments_api: Optional[EnvironmentsAPI] = None

    @property
    def rollout(self) -> RolloutAPI:
        if self._rollout_api is None:
            self._rollout_api = RolloutAPI(
                send_loop_config=self.config.send_loop_config, 
                shutdown_timeout=self.config.shutdown_timeout, 
                process_name=self.config.process_name
            )
        assert self._rollout_api is not None # helping the type checker
        return self._rollout_api

    @property
    def environments(self) -> EnvironmentsAPI:
        if self._environments_api is None:
            self._environments_api = EnvironmentsAPI(
                base_url=self.environments_base_url,
                api_key=self.api_key
            )
        assert self._environments_api is not None # helping the type checker
        return self._environments_api


    def sandbox(
        self, 
        settings: SandboxSettings,
        creation_timeout: int = 60*30,
    ) -> SandboxesAPI:
        if self.api_key is None:
            raise ValueError("API key is required for sandbox API")

        return SandboxesAPI(
            base_url=self.sandboxes_base_url,
            api_key=self.api_key,
            settings=settings,
            creation_timeout=creation_timeout,
        )


import inspect
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Any, Awaitable, Callable, Sequence, TypeVar, get_type_hints

from pydantic import BaseModel, ValidationError

from .types import (Blocks, JSONObject, ListToolsOutput, RunToolError,
                    RunToolOutput, RunToolSuccess, ToolOutput, ToolSpec)
from .utils import maybe_await

T = TypeVar("T")

logger = getLogger(__file__)


def tool(fn: Callable[..., Any]) -> Callable[..., Any]:
    setattr(fn, "_env_tool", True)
    return fn

class Environment(ABC):
    """
    An environment is an interface to computation that may be stateful. Clients interface with the
    environment through a persistent connection to perform actions. Environments have _tasks_, which
    are JSON objects describing a particular setup and goal state. For example, inside of an Ubuntu
    environment, a task could be to download a file from the internet and save its contents to a csv
    file.
    """
    def __init__(self, task_spec: JSONObject = {}, secrets: dict[str, str] = {}) -> None:
        self.task_spec = task_spec

    def setup(self) -> None | Awaitable[None]:
        """
        Setup the environment. This is called upon the first tool call by a connected client.
        """
        pass

    def teardown(self) -> None | Awaitable[None]:
        """
        Teardown the environment. This is called upon client disconnect.
        """
        pass

    @abstractmethod
    def get_prompt(self) -> Blocks | Awaitable[Blocks]:
        """
        Get a default prompt for the current task. For example, if the task is a question-answer pair,
        returning the question would be a sensible choice here.
        """

    @classmethod
    @abstractmethod
    def list_tasks(cls, split: str) -> Sequence[JSONObject]:
        """
        Get a list of tasks for the given split. Default is the empty list.
        """

    @classmethod
    @abstractmethod
    def list_splits(cls) -> list[str]:
        """
        Get a list of splits for the environment. Default is the empty list.
        """

    @staticmethod
    def _is_tool(fn: Callable[..., Any]) -> bool:
        if not callable(fn) or not getattr(fn, "_env_tool", False):
            return False
        real = inspect.unwrap(fn)
        hints = get_type_hints(real, include_extras=True)
        params = [p for p in inspect.signature(real).parameters.values() if p.name != "self"]
        ret = hints.get("return")
        if len(params) == 0:
            return ret == ToolOutput
        if len(params) == 1:
            pt = hints.get(params[0].name)
            return (
                pt is not None and ret is not None and inspect.isclass(pt)
                and issubclass(pt, BaseModel) and ret == ToolOutput
            )
        return False

    @classmethod
    def list_tools(cls) -> ListToolsOutput:
        out: list[ToolSpec] = []
        for name in dir(cls):
            fn = getattr(cls, name)
            if not cls._is_tool(fn):
                continue
            real = inspect.unwrap(fn)
            hints = get_type_hints(real, include_extras=True)
            params = [p for p in inspect.signature(real).parameters.values() if p.name != "self"]
            schema = None
            if params:
                mdl: type[BaseModel] = hints[params[0].name]  # type: ignore[assignment]
                schema = mdl.model_json_schema() if hasattr(mdl, "model_json_schema") else mdl.schema()  # type: ignore[attr-defined]
            out.append(ToolSpec(name=name, description=(fn.__doc__ or "").strip(), input_schema=schema))
        return ListToolsOutput(tools=out)

    async def _call_tool(self, name: str, input: JSONObject) -> RunToolOutput:
        fn = getattr(self, name, None)
        if fn is None or not self._is_tool(fn):
            return RunToolOutput(RunToolError(error=f"{name!r} is not a valid tool"))
        real = inspect.unwrap(fn)
        hints = get_type_hints(real, include_extras=True)
        params = [p for p in inspect.signature(real).parameters.values() if p.name != "self"]
        if not params:
            res = await maybe_await(fn())
        else:
            mdl: type[BaseModel] = hints[params[0].name]  # type: ignore[assignment]
            try:
                inp = mdl(**input)
            except ValidationError as e:
                return RunToolOutput(RunToolError(error=f"Tool input validation error: {str(e.errors())}"))
            res = await maybe_await(fn(inp))
        if not isinstance(res, ToolOutput):
            raise TypeError(f"{name!r} returned {type(res).__name__}; expected ToolOutput")
        return RunToolOutput(RunToolSuccess(output=res))

    @classmethod
    def name(cls) -> str:
        return cls.__name__


import asyncio
import base64
import os
import uuid
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Coroutine, overload

import aiohttp
from .http import request_retryable, resumable_sse
from .ping import ErrorResponse, PingManager
from .types import (
    ImageBlock,
    JSONObject,
    JSONValue,
    Provider,
    Server,
    Task,
    TextBlock,
    ToolCallError,
    ToolOutput,
    ToolSpec,
    Literal,
    Mapping,
)


def _strip_titles(value: Any) -> Any:
    """Recursively remove JSON schema `title` keys."""
    if isinstance(value, dict):
        return {
            k: _strip_titles(v)
            for k, v in value.items()
            if k != "title"
        }
    if isinstance(value, list):
        return [_strip_titles(item) for item in value]
    return value

@overload
def convert_tool_response(res: Mapping[str, Any], format: None = None) -> list[ToolSpec]: ...

@overload
def convert_tool_response(res: Mapping[str, Any], format: Provider = ...) -> list[dict[str, Any]]: ...

def convert_tool_response(
    res: Mapping[str, Any],
    format: Provider | None = None,
) -> list[ToolSpec] | list[dict[str, Any]]:
    if format is not None:
        if format == "openai":
            return [
                {
                    "type": "function",
                    **{
                        k: _strip_titles(v)
                        for k, v in tool.items()
                        if k not in {"input_schema", "title"}
                    },
                    "parameters": _strip_titles(tool["input_schema"]) if tool.get("input_schema") else None
                }
                for tool in res["tools"]
            ]
        elif format == "openrouter":
            return [
                {
                    "type": "function",
                    "function": {
                        k: _strip_titles(v)
                        for k, v in tool.items()
                        if k not in {"input_schema", "title"}
                    },
                    "parameters": _strip_titles(tool["input_schema"]) if tool.get("input_schema") else None
                }
                for tool in res["tools"]
            ]
        elif format == "anthropic":
            return [
                {
                    "type": "custom",
                    **{
                        k: _strip_titles(v)
                        for k, v in tool.items()
                        if k not in {"input_schema", "title"}
                    },
                    "input_schema": _strip_titles(tool["input_schema"]) if tool.get("input_schema") else None
                }
                for tool in res["tools"]
            ]
        elif format == "google":
            return [
                {
                    **{
                        k: _strip_titles(v)
                        for k, v in tool.items()
                        if k not in {"input_schema", "title"}
                    },
                    "parameters": _strip_titles(tool["input_schema"]) if tool.get("input_schema") else None
                }
                for tool in res["tools"]
            ]
        else:
            raise ValueError(f"Invalid format: {format!r}")

    return [ToolSpec(**tool) for tool in res["tools"]]

@asynccontextmanager
async def matrix_sid_provider(client: aiohttp.ClientSession, server_name: str, token: str | None) -> AsyncGenerator[str, None]:
    sid = await request_retryable(client, "POST", "/create_session", expect_json=True, deployment=server_name, token=token)
    try:
        yield sid["sid"]
    finally:
        await request_retryable(client, "POST", "/delete_session", sid=sid["sid"], expect_json=False, deployment=server_name, token=token)

def _finalize_session(session: aiohttp.ClientSession):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        try:
            asyncio.run(session.close())
        except RuntimeError:
            loop = asyncio.new_event_loop()
            loop.run_until_complete(session.close())
            loop.close()
    else:
        if not session.closed:
            loop.create_task(session.close())

class SessionTerminatedError(RuntimeError):
    def __init__(self, reason: str, *, sid: str | None):
        super().__init__(f"Session terminated (sid={sid!r}): {reason}")
        self.reason = reason
        self.sid = sid

class Session:
    
    def __init__(
        self, 
        env: "Environment", 
        task: Task, 
        secrets: dict[str, str] | None = None, 
        api_key: str | None = None
    ):
        self.client = env.client
        self.task = task
        self.api_key = api_key
        self.base_url = str(env.client._base_url)
        self._sid_cm = None
        self.sid: str | None = None

        self.secrets = {**(secrets or {}), **{"api_key": api_key}}
        
        self._ping_manager = env.ping_manager
        self._ping_id = uuid.uuid4().hex
        self._dead = asyncio.Event()
        self._dead_exception: SessionTerminatedError | None = None

    def _mark_dead(self, exc: ErrorResponse):
        if self._dead_exception is None:
            self._dead_exception = SessionTerminatedError(exc.message, sid=self.sid)
            self._dead.set()

    async def _run_task(self, coro: Coroutine[Any, Any, Any]):
        """Run a coroutine until completion or until the session dies."""
        if self._dead_exception is not None:
            raise self._dead_exception

        task = asyncio.create_task(coro)
        stopper = asyncio.create_task(self._dead.wait())
        try:
            done, pending = await asyncio.wait(
                {task, stopper},
                return_when=asyncio.FIRST_COMPLETED,
            )

            # If session died first
            if self._dead.is_set() and not task.done():
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass
                if self._dead_exception:
                    raise self._dead_exception

            return await task  # return result (or raise from task)

        finally:
            stopper.cancel()
            if not task.done():
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass

    async def __aenter__(self) -> "Session":
        self._sid_cm = matrix_sid_provider(self.client, self.task.deployment_name, self.api_key)
        self.sid = await self._sid_cm.__aenter__()
        await request_retryable(
            self.client,
            "POST",
            "/create",
            expect_json=True,
            sid=self.sid,
            deployment=self.task.deployment_name,
            json={
                "env_name": self.task.environment_name,
                "task_spec": self.task.task_spec,
                "secrets": self.secrets
            },
            token=self.api_key
        )

        # register with ping manager
        await self._ping_manager.start_ping(
            task_id=str(self._ping_id),
            url=f"{self.base_url}/ping",
            deployment=self.task.deployment_name,
            session_id=self.sid,
            api_key=self.api_key,
            sleep_time=10,
            on_error=self._mark_dead,
        )

        return self

    async def __aexit__(self, *exc):
        try:
            await self._ping_manager.stop_ping(str(self._ping_id))
        except:
            pass
        
        try:
            await request_retryable(
                self.client,
                "POST",
                "/delete",
                expect_json=False,
                sid=self.sid,
                deployment=self.task.deployment_name,
                token=self.api_key
            )
        except:
            pass
        await self._sid_cm.__aexit__(*exc) # type: ignore
        self._sid_cm = None

    async def get_prompt(self) -> list[TextBlock | ImageBlock]:
        res = await self._run_task(
            request_retryable(
                self.client,
                "GET",
                f"/{self.task.environment_name}/prompt",
                expect_json=True,
                sid=self.sid,
                deployment=self.task.deployment_name,
                token=self.api_key,
            )
        )
        blocks: list[TextBlock | ImageBlock] = []
        for block in res:
            if block["type"] == "text":
                blocks.append(TextBlock(text=block["text"], detail=block["detail"]))
            elif block["type"] == "image":
                blocks.append(ImageBlock(mimeType=block["mimeType"], detail=block["detail"], data=block["data"]))
        return blocks

    @overload
    async def list_tools(self, format: None = None) -> list[ToolSpec]: ...

    @overload
    async def list_tools(self, format: Provider) -> list[dict]: ...

    async def list_tools(self, format: Provider | None = None) -> list[ToolSpec] | list[dict]:
        res = await self._run_task(
            request_retryable(
                self.client,
                "GET",
                f"/{self.task.environment_name}/tools",
                expect_json=True,
                sid=self.sid,
                deployment=self.task.deployment_name,
                token=self.api_key,
            )
        )
        return convert_tool_response(res, format=format)

    async def call_tool(self, tool_name: str, input: JSONObject = {}) -> ToolOutput:
        if not isinstance(input, Mapping):
            raise ToolCallError(f"Tool input must be a dictionary, got {type(input).__name__}")

        if not all(isinstance(k, str) for k in input.keys()):
            non_string_keys = [k for k in input.keys() if not isinstance(k, str)]
            raise ToolCallError(f"All keys in tool input must be strings. Found non-string keys: {non_string_keys}")

        res = await self._run_task(
            resumable_sse(
                self.client,
                f"/{self.task.environment_name}/call",
                sid=self.sid,
                deployment=self.task.deployment_name,
                token=self.api_key,
                json={"name": tool_name, "input": input},
                max_retries=5,
            )
        )

        if res["ok"]:
            blocks: list[TextBlock | ImageBlock] = []
            for block in res["output"]["blocks"]:
                if block["type"] == "text":
                    blocks.append(TextBlock(
                        text=block["text"],
                        detail=block["detail"]
                    ))
                elif block["type"] == "image":
                    blocks.append(ImageBlock(
                        mimeType=block["mimeType"],
                        detail=block["detail"],
                        data=block["data"]
                    ))
            return ToolOutput(
                blocks=blocks,
                metadata=res["output"]["metadata"],
                reward=res["output"]["reward"],
                finished=res["output"]["finished"]
            )
        else:
            raise ToolCallError(res["error"])
        
class Environment:

    def __init__(
        self,
        namespace: str | None,
        name: str,
        variant: str | None,
        client: aiohttp.ClientSession,
        api_key: str | None,
        ping_manager: PingManager
    ) -> None:

        self.server = name
        self.namespace = namespace
        self.name = name
        self.variant = variant
        self.client = client
        self.api_key = api_key
        self.ping_manager = ping_manager

    @property
    def deployment_name(self) -> str:
        if self.namespace is None:
            return self.name
        else:
            return f"{self.namespace}/{self.name}"
        
    async def list_splits(self) -> list[str]:
        async with matrix_sid_provider(self.client, self.deployment_name, self.api_key) as sid:
            path = "/splits" if self.variant is None else f"/{self.variant}/splits"
            res = await request_retryable(self.client, "GET", path, expect_json=True, sid=sid, deployment=self.deployment_name, token=self.api_key)
            return res

    async def list_tasks(self, split: str) -> list[Task]:
        async with matrix_sid_provider(self.client, self.deployment_name, self.api_key) as sid:
            path = "/tasks" if self.variant is None else f"/{self.variant}/tasks"
            res = await request_retryable(self.client, "POST", path, expect_json=True, sid=sid, deployment=self.deployment_name, json={"split": split}, token=self.api_key)
            return [Task(server_name=self.server, environment_name=res["env_name"], task_spec=task, namespace=self.namespace) for task in res["tasks"]]
        
    async def list_tools(self, format: Provider | None = None) -> list[ToolSpec] | list[dict]:
        path = "/tools" if self.variant is None else f"/{self.variant}/tools"
        async with matrix_sid_provider(self.client, self.deployment_name, self.api_key) as sid:
            res = await request_retryable(self.client, "GET", path, expect_json=True, sid=sid, deployment=self.deployment_name, token=self.api_key)
            return convert_tool_response(res, format=format)
        
    async def get_prompt(self, task: Task) -> str:
        async with matrix_sid_provider(self.client, task.deployment_name, self.api_key) as sid:
            path = "/prompt" if self.variant is None else f"/{self.variant}/prompt"
            res = await request_retryable(self.client, "GET", path, expect_json=True, sid=sid, deployment=task.deployment_name, token=self.api_key)
            return res
        
    def session(self, task: Task, secrets: dict[str, str] | None = None) -> Session:
        return Session(self, task, secrets, self.api_key)

class EnvironmentsAPI:

    def __init__(
        self,
        base_url: str,
        api_key: str
    ):
        self.api_key = api_key
        self.ping_manager = PingManager()

        self.base_url = base_url
        self.timeout = aiohttp.ClientTimeout(total=None)
        self.connector = aiohttp.TCPConnector(limit=1_000_000)

        self._clients: dict[str, aiohttp.ClientSession] = {}

    def get(self, name: str, variant: str | None = None, base_url: str | None = None) -> Environment:

        parts = name.split("/", maxsplit=1)
        namespace = None
        if len(parts) == 1:
            env_name = parts[0]
        elif len(parts) == 2:
            namespace, env_name = parts
            pass
        else:
            raise RuntimeError("impossible")


        if namespace and self.api_key is None:
            raise ValueError(f"Expected api_key to be passed when accessing remote environment")

        if base_url is None:
            base_url = self.base_url

        if base_url not in self._clients:
            self._clients[base_url] = aiohttp.ClientSession(base_url=base_url, timeout=self.timeout, connector=self.connector)
        client = self._clients[base_url]

        return Environment(
            namespace=namespace,
            name=env_name,
            variant=variant,
            client=client,
            api_key=self.api_key,
            ping_manager=self.ping_manager
        )

    def __del__(self):
        for client in self._clients.values():
            _finalize_session(client)
        self.ping_manager.shutdown()


import asyncio, aiohttp
import json as json_lib
from typing import Any, Optional, AsyncGenerator, Tuple, Dict
from tenacity import retry, stop_after_attempt, wait_exponential, retry_if_exception
from logging import getLogger

logger = getLogger("openreward-environments-client")

def _is_retryable_http_error(exception: BaseException) -> bool:
    if isinstance(exception, (aiohttp.ClientError, asyncio.TimeoutError)):
        return True
    if isinstance(exception, aiohttp.ClientResponseError):
        return exception.status >= 500
    return False

@retry(
    retry=retry_if_exception(_is_retryable_http_error),
    wait=wait_exponential(multiplier=1, min=1, max=10),
    stop=stop_after_attempt(3),
    reraise=True,
)
async def request_retryable(
    client: aiohttp.ClientSession,
    method: str,
    path: str,
    expect_json: bool,
    token: str | None,
    json: dict[str, Any] | None = None,
    sid: str | None = None,
    deployment: str | None = None,
) -> Any:
    headers: dict[str, str] = {}
    if token is not None:
        headers["Authorization"] = f"Bearer {token}"
        headers["X-API-Key"] = token
    if sid:
        headers["X-Session-ID"] = sid
    if deployment:
        headers["X-Deployment"] = deployment

    async with client.request(method, path, headers=headers, json=json) as response:
        text = await response.text()  # read body first
        try:
            response.raise_for_status()
        except aiohttp.ClientResponseError as e:
            raise aiohttp.ClientResponseError(
                request_info=e.request_info,
                history=e.history,
                status=e.status,
                message=text,
                headers=e.headers,
            )
        return await response.json() if expect_json else None

class HeartbeatTimeoutError(Exception):
    pass

class MaxRetriesError(Exception):
    pass

async def _parse_sse_events(
    response: aiohttp.ClientResponse,
) -> AsyncGenerator[Tuple[str, str], None]:
    """Parses an aiohttp response stream and yields SSE events."""
    event = None
    data_lines: list[str] = []
    async for raw_line in response.content:
        line = raw_line.decode("utf-8", "ignore").rstrip("\r\n")

        if not line:
            if event:
                yield event, "\n".join(data_lines)
            event = None
            data_lines = []
            continue

        if line.startswith(":"):
            continue

        field, value = line.split(":", 1)
        value = value.lstrip()

        if field == "event":
            event = value
        elif field == "data":
            data_lines.append(value)

# The simplified main function
async def resumable_sse(
    client: aiohttp.ClientSession,
    path: str,
    token: str | None,
    *,
    json: Optional[Dict[str, Any]] = None,
    sid: Optional[str] = None,
    deployment: Optional[str] = None,
    task_id: Optional[str] = None,
    max_retries: Optional[int] = None,
    backoff_base: float = 0.5,
    backoff_max: float = 10.0,
    timeout: Optional[float] = None,
    heartbeat_timeout: int = 30
) -> Any:

    client_timeout = aiohttp.ClientTimeout(total=None, sock_read=heartbeat_timeout)
    payload = dict(json or {})
    headers: dict[str, str] = {
        "Accept": "text/event-stream"
    }
    if token:
        headers["Authorization"] = f"Bearer {token}"
        headers["X-API-Key"] = token
    if sid:
        headers["X-Session-ID"] = sid
    if deployment:
        headers["X-Deployment"] = deployment

    async def _execute_with_retries():
        nonlocal task_id, payload
        attempt = 0
        while True:
            if task_id:
                payload["task_id"] = task_id

            try:
                async with client.post(path, headers=headers, json=payload, timeout=client_timeout) as resp:
                    resp.raise_for_status()
                    attempt = 0

                    chunks = []
                    async for event, data in _parse_sse_events(resp):
                        if event == "task_id":
                            task_id = data.strip()
                        elif event == "chunk":
                            chunks.append(data)
                        elif event == "end":
                            chunks.append(data)
                            final_result = "".join(chunks)
                            return json_lib.loads(final_result)
                        elif event == "error":
                            raise RuntimeError(data or "Unknown SSE error")

                    raise aiohttp.ClientPayloadError("Stream ended unexpectedly")

            except aiohttp.ClientResponseError as e:
                if e.status < 500:
                    raise e

            except aiohttp.ClientError as e:
                logger.warning(f"Client error: {e}")
                pass

            except RuntimeError as e:
                raise e

            except asyncio.TimeoutError:
                raise HeartbeatTimeoutError()

            attempt += 1
            if max_retries is not None and attempt > max_retries:
                raise MaxRetriesError("Exceeded max retries for a retryable error.")

            delay = min(backoff_max, backoff_base * (2 ** (attempt - 1)))
            await asyncio.sleep(delay)


    try:
        return await asyncio.wait_for(_execute_with_retries(), timeout=timeout)
    except asyncio.TimeoutError:
        raise TimeoutError(f"Total operation timed out after {timeout} seconds.") from None


import pytest
import asyncio
import uvicorn
import aiohttp
from threading import Thread
from typing import Generator
from openreward import OpenReward
from openreward.environments import Environment, Server, tool, ToolOutput
from openreward.environments.types import Blocks, TextBlock, JSONObject


class Foo(Environment):
    def setup(self):
        pass

    def teardown(self):
        pass

    def get_prompt(self) -> Blocks:
        return [TextBlock(text=str(self.task_spec["foo"]))]

    @classmethod
    def list_splits(cls) -> list[str]:
        return ["train"]

    @classmethod
    def list_tasks(cls, split: str) -> list[JSONObject]:
        assert split == "train"
        return [{"foo": "bar"}]

    @tool
    async def submit(self) -> ToolOutput:
        return ToolOutput(blocks=[TextBlock(text="foo_result")], reward=1.0, finished=True)


class Bar(Environment):
    def setup(self):
        pass

    def teardown(self):
        pass

    def get_prompt(self) -> Blocks:
        return [TextBlock(text=str(self.task_spec["bar"]))]

    @classmethod
    def list_splits(cls) -> list[str]:
        return ["test"]

    @classmethod
    def list_tasks(cls, split: str) -> list[JSONObject]:
        assert split == "test"
        return [{"bar": "baz"}]

    @tool
    async def submit(self) -> ToolOutput:
        return ToolOutput(blocks=[TextBlock(text="bar_result")], reward=0.5, finished=True)


async def wait_for_server(base_url: str, timeout: float = 5.0):
    """Wait for server to be ready using aiohttp."""
    import time
    start = time.monotonic()
    async with aiohttp.ClientSession() as session:
        while time.monotonic() - start < timeout:
            try:
                async with session.get(f"{base_url}/health", timeout=aiohttp.ClientTimeout(total=0.5)) as resp:
                    if resp.status == 200:
                        return
            except Exception:
                pass
            await asyncio.sleep(0.1)
    pytest.fail("Server failed to start")


@pytest.fixture(scope="module")
def server() -> Generator[str, None, None]:
    """Start the server in a background thread and yield the base URL."""
    import os
    os.environ["OPENREWARD_LOCAL"] = "1"
    
    host = "localhost"
    port = 8080
    app = Server(environments=[Foo, Bar]).app
    config = uvicorn.Config(app, host=host, port=port, log_level="warning")
    server_instance = uvicorn.Server(config)
    
    thread = Thread(target=server_instance.run, daemon=True)
    thread.start()
    
    base_url = f"http://{host}:{port}"
    asyncio.run(wait_for_server(base_url))
    
    yield base_url
    server_instance.should_exit = True


@pytest.fixture
def client(server: str) -> OpenReward:
    return OpenReward(base_url=server, environments_base_url=server, api_key="test")


# =============================================================================
# Tests for default environment (no variant specified, uses first env)
# =============================================================================

@pytest.mark.asyncio
async def test_default_variant_splits(client: OpenReward):
    """Test that default variant works - redirects to first environment."""
    environment = client.environments.get("foo")  # No variant specified
    splits = await environment.list_splits()
    assert splits == ["train"]


@pytest.mark.asyncio
async def test_default_variant_tools(client: OpenReward):
    """Test listing tools with default variant."""
    environment = client.environments.get("foo")
    tools = await environment.list_tools()
    tool_names = [t.name for t in tools]
    assert "submit" in tool_names


@pytest.mark.asyncio
async def test_default_variant_list_tasks(client: OpenReward):
    """Test listing tasks with default variant."""
    environment = client.environments.get("foo")
    tasks = await environment.list_tasks(split="train")
    assert len(tasks) == 1
    assert tasks[0].task_spec == {"foo": "bar"}


@pytest.mark.asyncio
async def test_default_variant_call_tool(client: OpenReward):
    """Test calling tool with default variant."""
    environment = client.environments.get("foo")
    tasks = await environment.list_tasks(split="train")
    async with environment.session(tasks[0]) as session:
        res = await session.call_tool("submit")
        assert res.reward == 1.0
        assert res.finished is True
        assert len(res.blocks) == 1
        assert res.blocks[0].type == "text"
        assert res.blocks[0].text == "foo_result"


# =============================================================================
# Tests for explicit variants
# =============================================================================

@pytest.mark.asyncio
async def test_explicit_variant_foo_splits(client: OpenReward):
    """Test Foo environment with explicit variant."""
    environment = client.environments.get("foo", variant="foo")
    splits = await environment.list_splits()
    assert splits == ["train"]


@pytest.mark.asyncio
async def test_explicit_variant_bar_splits(client: OpenReward):
    """Test Bar environment with explicit variant."""
    environment = client.environments.get("bar", variant="bar")
    splits = await environment.list_splits()
    assert splits == ["test"]


@pytest.mark.asyncio
async def test_explicit_variant_bar_tools(client: OpenReward):
    """Test Bar environment tools with explicit variant."""
    environment = client.environments.get("bar", variant="bar")
    tools = await environment.list_tools()
    tool_names = [t.name for t in tools]
    assert "submit" in tool_names


@pytest.mark.asyncio
async def test_explicit_variant_bar_list_tasks(client: OpenReward):
    """Test Bar environment task listing with explicit variant."""
    environment = client.environments.get("bar", variant="bar")
    tasks = await environment.list_tasks(split="test")
    assert len(tasks) == 1
    assert tasks[0].task_spec == {"bar": "baz"}


@pytest.mark.asyncio
async def test_explicit_variant_bar_call_tool(client: OpenReward):
    """Test Bar environment tool call with explicit variant."""
    environment = client.environments.get("bar", variant="bar")
    tasks = await environment.list_tasks(split="test")
    async with environment.session(tasks[0]) as session:
        res = await session.call_tool("submit")
        assert res.reward == 0.5
        assert res.finished is True
        assert len(res.blocks) == 1
        assert res.blocks[0].type == "text"
        assert res.blocks[0].text == "bar_result"


# =============================================================================
# Tests for multiple variants interaction
# =============================================================================

@pytest.mark.asyncio
async def test_multiple_variants_different_splits(client: OpenReward):
    """Test that different variants return different splits."""
    foo_env = client.environments.get("foo", variant="foo")
    bar_env = client.environments.get("bar", variant="bar")
    
    foo_splits = await foo_env.list_splits()
    bar_splits = await bar_env.list_splits()
    
    assert foo_splits == ["train"]
    assert bar_splits == ["test"]
    assert foo_splits != bar_splits


@pytest.mark.asyncio
async def test_multiple_variants_different_tasks(client: OpenReward):
    """Test that different variants return different tasks."""
    foo_env = client.environments.get("foo", variant="foo")
    bar_env = client.environments.get("bar", variant="bar")
    
    foo_tasks = await foo_env.list_tasks(split="train")
    bar_tasks = await bar_env.list_tasks(split="test")
    
    assert foo_tasks[0].task_spec == {"foo": "bar"}
    assert bar_tasks[0].task_spec == {"bar": "baz"}


@pytest.mark.asyncio
async def test_multiple_variants_concurrent_sessions(client: OpenReward):
    """Test running sessions on multiple variants concurrently."""
    foo_env = client.environments.get("foo", variant="foo")
    bar_env = client.environments.get("bar", variant="bar")
    
    foo_tasks = await foo_env.list_tasks(split="train")
    bar_tasks = await bar_env.list_tasks(split="test")
    
    async with foo_env.session(foo_tasks[0]) as foo_session:
        async with bar_env.session(bar_tasks[0]) as bar_session:
            foo_res = await foo_session.call_tool("submit")
            bar_res = await bar_session.call_tool("submit")
            
            assert foo_res.reward == 1.0
            assert foo_res.blocks[0].text == "foo_result"
            
            assert bar_res.reward == 0.5
            assert bar_res.blocks[0].text == "bar_result"


@pytest.mark.asyncio
async def test_multiple_variants_prompt_isolation(client: OpenReward):
    """Test that prompts are correctly isolated between variants."""
    foo_env = client.environments.get("foo", variant="foo")
    bar_env = client.environments.get("bar", variant="bar")
    
    foo_tasks = await foo_env.list_tasks(split="train")
    bar_tasks = await bar_env.list_tasks(split="test")
    
    async with foo_env.session(foo_tasks[0]) as foo_session:
        foo_prompt = await foo_session.get_prompt()
        assert foo_prompt[0].text == "bar"  # from {"foo": "bar"}
    
    async with bar_env.session(bar_tasks[0]) as bar_session:
        bar_prompt = await bar_session.get_prompt()
        assert bar_prompt[0].text == "baz"  # from {"bar": "baz"}


from .api.rollouts.rollout import Rollout
from .client import OpenReward, RolloutAPI
from .api.rollouts.serializers.base import (
    AssistantMessage,
    ReasoningItem,
    SystemMessage,
    ToolCall,
    ToolResult,
    UploadType,
    UserMessage,
)
from .api.sandboxes import SandboxSettings, SandboxBucketConfig, SandboxesAPI

__all__ = ["OpenReward", "Rollout", "RolloutAPI", "UserMessage", "AssistantMessage", "SystemMessage", "ReasoningItem", "ToolCall", "ToolResult", "UploadType", "SandboxSettings", "SandboxBucketConfig", "SandboxesAPI"]



from typing import Literal, Optional

from pydantic import BaseModel, field_validator

# Machine size in format 'cpu:memory' (e.g., '1:2' = 1 CPU, 2GB memory)
MachineSize = Literal[
    '0.1:0.1',
    '0.5:0.5',
    '1:1',
    '2:2',
    '4:4',
    '0.1:0.2',
    '0.5:1',
    '1:2',
    '2:4',
    '4:8',
    '0.1:0.4',
    '0.5:2',
    '1:4',
    '2:8',
    '4:16',
]

class SandboxBucketConfig(BaseModel):
    mount_path: str
    """Path inside the container where the bucket will be mounted."""

    read_only: bool = True
    """Whether the bucket should be mounted in read-only mode."""

    only_dir: Optional[str] = None
    """If set, only mount the specified directory from the bucket."""

    implicit_dirs: bool = True
    """If True, mount all subdirectories of the bucket."""

class SandboxSidecarContainer(BaseModel):
    name: str
    """Unique name for this sidecar container."""

    image: str
    """Container image to run."""

    command: Optional[list[str]] = None
    """Optional command override for the container entrypoint."""

    args: Optional[list[str]] = None
    """Optional arguments to pass to the command."""

    env: Optional[dict[str, str]] = None
    """Environment variables for this container."""

    machine_size: MachineSize
    """Machine size to run the container in."""

    ports: Optional[list[int]] = None
    """Ports this container exposes (informational, used for probes)."""

    @field_validator("name")
    def validate_name_not_reserved(cls, value: str) -> str:
        if value.lower() in {"main", "sidecar"}:
            raise ValueError("Sidecar container names 'main' and 'sidecar' are reserved.")
        return value


class SandboxHostAlias(BaseModel):
    """Configuration for adding hostname aliases to /etc/hosts."""

    ip: str
    """IP address to map hostnames to, e.g. '127.0.0.1'."""

    hostnames: list[str]
    """List of hostnames to map to the IP"""

class SandboxSettings(BaseModel):
    """Compute resource settings for a container."""

    environment: str
    """OpenReward environment to run the container in."""

    image: str
    """Container image to run, e.g. "python:3.10-slim"."""

    machine_size: MachineSize
    """
    Machine size to run the container in.
    Example: "0.1:0.2" = 0.1 CPU and 0.2 GB memory.
    """

    # TODO: do we want these?
    # disk_request: Optional[str] = None
    # """Minimum ephemeral storage requested (if using emptyDir volumes). Example: "1Gi" = 1 GB."""

    # disk_limit: Optional[str] = None
    # """Maximum ephemeral storage allowed; exceeding this causes eviction."""

    env: Optional[dict[str, str]] = None
    """Environment variables for the container, e.g. {"ENV": "prod"}."""

    block_network: bool = False
    """If True, disables outbound network access (Kubernetes egress policies required)."""

    bucket_config: Optional[SandboxBucketConfig] = None
    """List of buckets to mount; mounts them into the container at runtime."""

    # labels: Optional[dict[str, str]] = None
    # """Labels to add to running computer for attribution."""

    sidecars: Optional[list[SandboxSidecarContainer]] = None
    """Additional containers to run in the same pod."""

    host_aliases: Optional[list[SandboxHostAlias]] = None
    """Hostname aliases to add to /etc/hosts in all containers."""

class PodTerminatedError(RuntimeError):
    def __init__(self, reason: str, *, client_id: str | None):
        super().__init__(f"Pod terminated (client_id={client_id!r}): {reason}")
        self.reason = reason
        self.client_id = client_id



import asyncio
import base64
from pathlib import Path
from typing import Tuple

import aiohttp
from openreward.api.sandboxes.http import request_retryable, resumable_sse
from openreward.api.sandboxes.ping import ErrorResponse, get_ping_manager
from openreward.api.sandboxes.types import PodTerminatedError, SandboxSettings


def _decode_output(output: str) -> str:
    """Output from the terminal is base64 encoded, as it can arbitrary binary data."""
    return base64.b64decode(output.encode('utf-8')).decode('utf-8', 'surrogateescape').rstrip()

def _finalize_session(session: aiohttp.ClientSession):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        try:
            asyncio.run(session.close())
        except RuntimeError:
            loop = asyncio.new_event_loop()
            loop.run_until_complete(session.close())
            loop.close()
    else:
        if not session.closed:
            loop.create_task(session.close())

class SandboxesAPI:
    def __init__(self,
        base_url: str,
        api_key: str,
        settings: SandboxSettings,
        creation_timeout: int = 60*30,
        # session_headers = None,
    ) -> None:
        self.base_url = base_url
        self.api_key = api_key
        self.settings = settings
        self.creation_timeout = creation_timeout
        
        # TODO: maybe use shared client session for all sandboxes?
        self.connector = aiohttp.TCPConnector(limit=1_000_000)
        # self.client = aiohttp.ClientSession(base_url=base_url, connector=self.connector, headers=self.session_headers)
        self.client = aiohttp.ClientSession(base_url=base_url, connector=self.connector)
        self.client_id = None

        self._ping_manager = get_ping_manager()
        self._ping_id = id(self)  # Unique ID for this sandbox
        self._dead = asyncio.Event()
        self._dead_exception: BaseException | None = None

    def _ensure_alive(self):
        if self._dead_exception is not None:
            raise self._dead_exception

    def _mark_dead(self, exc: ErrorResponse):
        if self._dead_exception is None:
            self._dead_exception = PodTerminatedError(exc.message, client_id=self.client_id)
            self._dead.set()

    async def run(
        self,
        cmd: str,
        timeout: float | None = 300,
        max_bytes: int | None = 10_000_000, # 10mb
    ) -> Tuple[str, int]:
        """Run a command in the container."""
        self._ensure_alive()

        run_task = asyncio.create_task(resumable_sse(
            self.client,
            "/run",
            token=self.api_key,
            json={
                "cmd": cmd,
                "timeout_s": timeout,
                "max_bytes": max_bytes,
                "shell": "/bin/bash",
            },
            client_id=self.client_id,
            max_retries=5,
        ))
        dead_task = asyncio.create_task(self._dead.wait())

        done, pending = await asyncio.wait(
            [run_task, dead_task],
            return_when=asyncio.FIRST_COMPLETED,
        )

        if dead_task in done:
            run_task.cancel()
            try:
                await run_task
            except asyncio.CancelledError:
                pass
            raise self._dead_exception or PodTerminatedError("Pod died", client_id=self.client_id)

        dead_task.cancel()
        try:
            res = await run_task
        finally:
            try:
                await dead_task
            except asyncio.CancelledError:
                pass

        return_code = res["return_code"]
        output = _decode_output(res["output"])
        return output, return_code

    async def check_run(
        self,
        cmd: str,
        timeout: float | None = 300,
        max_bytes: int | None = 10_000_000, # 10mb
    ) -> str:
        """Run a command in the container and raise an error if it fails."""
        self._ensure_alive()
        output, exit_code = await self.run(cmd, timeout=timeout, max_bytes=max_bytes)
        if exit_code != 0:
            raise RuntimeError(f"Command failed: {cmd}\n{output}")
        return output

    # TODO: nicer upload and download methods on the sandbox server
    async def upload(self, local_path: str | Path, container_path: str) -> None:
        """Upload a single file from local filesystem to the container."""
        self._ensure_alive()
        local_path = Path(local_path)
        if not local_path.exists():
            raise FileNotFoundError(f"Local file not found: {local_path}")

        max_size = 10 * 1024 * 1024 # make sure the file is not too large (10mb)
        if local_path.stat().st_size > max_size:
            raise ValueError(f"File is too large: {local_path.stat().st_size} bytes > {max_size} bytes")

        file_content = local_path.read_bytes()
        encoded_content = base64.b64encode(file_content).decode('ascii')

        cmd = f"echo '{encoded_content}' | base64 -d > {container_path}"
        await self.check_run(cmd, max_bytes=max_size)


    async def download(self, container_path: str) -> bytes:
        """Download a single file from the container."""
        self._ensure_alive()
        cmd = f"base64 {container_path}"
        output = await self.check_run(cmd, max_bytes=None)

        try:
            file_content = base64.b64decode(output.encode('ascii'))
            return file_content
        except Exception as e:
            raise RuntimeError(f"Failed to decode and write file: {e}")

    async def start(self) -> None:
        # get client id
        res = await resumable_sse(
            self.client,
            "/create",
            token=self.api_key,
            json={
                "creation_request": self.settings.model_dump(),
            },
            max_retries=3,
            timeout=self.creation_timeout,
        )
        self.client_id = res["client_id"]

        # Register with ping manager
        await self._ping_manager.start_ping(
            task_id=str(self._ping_id),
            url=f"{self.base_url}/ping",
            client_id=self.client_id,
            api_key=self.api_key,
            sleep_time=10,
            on_error=self._mark_dead,
        )

    async def stop(self) -> None:
        # stop ping
        try:
            await self._ping_manager.stop_ping(str(self._ping_id))
        except:
            pass

        # delete pod
        try:
            await request_retryable(
                self.client,
                "POST",
                "/delete",
                expect_json=True,
                token=self.api_key,
                client_id=self.client_id,
            )
        except:
            pass
        # close client
        if not self.client.closed:
            await self.client.close()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, *_):
        await self.stop()


    def __del__(self):
        _finalize_session(self.client)


from typing import List

try:
    from typing import NotRequired
except ImportError:
    from typing_extensions import NotRequired

import json

from anthropic.types import MessageParam as AnthropicMessageParam

from .models import NormalizedEvent
from .utils import get


def serialize_anthropic_message(message: AnthropicMessageParam) -> List[NormalizedEvent]:
    events: List[NormalizedEvent] = []
    role = get(message, "role")
    content = get(message, "content")

    role_to_type = {
        "user": "user_message",
        "assistant": "assistant_message",
        "system": "system_message",
    }

    # Simple string content -> visible message
    if isinstance(content, str):
        events.append(NormalizedEvent(type=role_to_type.get(role), content=content))
    # List content -> iterate blocks
    elif isinstance(content, list):
        for block in content:
            btype = get(block, "type")

            # Visible text blocks
            if btype in (None, "text"):
                txt = get(block, "text") or get(block, "content")
                events.append(NormalizedEvent(type=role_to_type.get(role), content=txt))
            # Hidden reasoning
            elif btype == "thinking":
                signature = get(block, "signature")
                thinking_text = get(block, "thinking")
                cref = f"anthropic:thinking:{signature}" if signature else "anthropic:thinking"
                events.append(
                    NormalizedEvent(
                        type="reasoning",
                        content=thinking_text,
                        content_reference=cref,
                    )
                )
            # Redacted hidden reasoning
            elif btype == "redacted_thinking":
                thinking_text = get(block, "data")
                events.append(
                    NormalizedEvent(
                        type="reasoning",
                        content=thinking_text,
                    )
                )
            # Tool call
            elif btype == "tool_use":
                name = get(block, "name")
                tool_id = get(block, "id")
                tool_input = get(block, "input")
                args_str = json.dumps(tool_input or {}, ensure_ascii=False, separators=(",", ":"))
                events.append(
                    NormalizedEvent(
                        type="tool_call",
                        name=name,
                        content=args_str,
                        call_id=tool_id,
                    )
                )
            # Tool result (Anthropic delivers these in a 'user' role wrapper)
            elif btype == "tool_result":
                tool_use_id = get(block, "tool_use_id")
                block_content = get(block, "content")
                tr_content = ""
                if isinstance(block_content, str):
                    tr_content = block_content
                elif isinstance(block_content, list):
                    for part in block_content:
                        if part.get("type") == "text":
                            tr_content += f"{part.get('text')}\n"
                events.append(
                    NormalizedEvent(
                        type="tool_result",
                        content=tr_content.strip(),
                        call_id=tool_use_id,
                    )
                )
            else:
                raise ValueError(f"Unknown block type: {btype}")
    else:
        raise ValueError(f"Unknown content type: {type(content)}")

    return events

if __name__ == "__main__":
    from anthropic.types.beta import BetaThinkingBlock, BetaToolUseBlock
    raw = [{'content': "I'm at the Ferry Building in San Francisco. Use the tools to: "
                '(1) get current weather, and (2) suggest 2 good cafés nearby. '
                'Then give me a 1-paragraph plan.',
    'role': 'user'},
    {'content': [BetaThinkingBlock(signature='EtgECkYICRgCKkDcm4hfdHjRJyAuiQlOh5hrYZSl0dcL6wHrZeVtdCkKbeWATJtbn5/0/Ifj2A59gyPv/0HW0Mfgu2LSsWfmNX+YEgzXh6l/3hBzFU1wYI0aDOcAjbfFrx6miY9vXiIweQYY3luw8Jf0JXkAhp5QpxmNSLpGE2r/DPOqh5VYNbM+D1Sxka3v4WuEQIiczwK7Kr8Dsh3ViLRLECrs7xjlUuzfwSo9MSLe03tYn8HKrs7NKridXJRSIhjL8CX3u08IdPq1/jkflQAZ2Kuug7Ph5MOyITE9xh+yfpqekk1srpnupbPzUzbpUtLQJswkAYuuBqmbiF+zLJGegNZWAFYgkvKaYBx4HxGoe7gu3k3rUML6+5G5hdlQczOW22tgj/4x5G8zM3PLAYNHFALVkXMy8betyhr8HrDiAyY9+KqNkAmCvImTmpYXOtU4vmbJ4Fv9+Zp/MPw+iukhBeGzJ+KjeEOb65D+8/8fJJJP5fAh5R9vIfLPJY31XxA6nFGM/Xna6DzMT3dw27t9XGprJLOzA/E+IGDgGJ8l6/+BVj4VIr8kZJ1lcYeu3dWujQiciMUcjBi/gLLZLYb09x8Wh9HrYhWTLEXXZnsrbDyPItpglF665mhUBEg3ZFXTamY6lBKKuAyvZVOFuS6mfgNRFJB/CXo9jOFelx+mHcIYKMRgL2vGYaymGQeZnLSVKNwozYmneCqcW9aa5/8EmPPk4Ay7b8ZN4DifoVqpTIXijWPnyHXyp2zCZx8SDnPCW/EZtge+APhv4g9Fhw8SbuRx5HYlgIJ9GAE=', thinking='The user wants me to:\n1. Get current weather for San Francisco (they\'re at the Ferry Building)\n2. Find 2 cafés near the Ferry Building in San Francisco\n3. Provide a 1-paragraph plan based on this information\n\nI have both tools available and can call them simultaneously since they don\'t depend on each other. Let me make both calls:\n- get_weather for San Francisco\n- find_cafes near "Ferry Building, San Francisco" with limit of 2', type='thinking'),
                BetaToolUseBlock(id='toolu_01LGnrWwnnQSDY5V5pZDrh8m', input={'location': 'San Francisco'}, name='get_weather', type='tool_use'),
                BetaToolUseBlock(id='toolu_01MSkzmTN6GJqcSqzFAyfedH', input={'near': 'Ferry Building, San Francisco', 'limit': 2}, name='find_cafes', type='tool_use')],
    'role': 'assistant'},
    {'content': [{'content': '{"location": "San Francisco", "temp": 18, "unit": '
                            '"celsius", "condition": "Cloudy"}',
                'tool_use_id': 'toolu_01LGnrWwnnQSDY5V5pZDrh8m',
                'type': 'tool_result'},
                {'content': '{"near": "Ferry Building, San Francisco", '
                            '"results": ["Blue Bottle (66 Mint St)", "Sightglass '
                            '(270 7th St)"]}',
                'tool_use_id': 'toolu_01MSkzmTN6GJqcSqzFAyfedH',
                'type': 'tool_result'}],
    'role': 'user'}]

    for message in raw:
        normalized = serialize_anthropic_message(message)
        print(normalized)



import asyncio
import inspect
import logging
import time
import traceback
import uuid
from contextlib import asynccontextmanager, suppress

import uvicorn
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.responses import RedirectResponse
from fastapi.routing import APIRoute
from sse_starlette import EventSourceResponse

from .environment import Environment
from .reconnect import sse_task_stream
from .types import Blocks, CreateSession, ListTasks, ToolCall
from .utils import maybe_await


def _get_env_map(environments: list[type[Environment]]) -> dict[str, type[Environment]]:
    env_map = {}
    for env in environments:
        # Always treat env as a class
        cls = env if inspect.isclass(env) else type(env)
        # Ensure we don't pass non-type or overloads to issubclass
        if not (isinstance(cls, type) and issubclass(cls, Environment)):
            raise TypeError(f"{cls!r} is not Environment")
        key: str = cls.name().lower()
        if key in env_map:
            raise ValueError(f"duplicate env {key}")
        env_map[key] = cls
    return env_map

def _get_env_cls(env_map: dict[str, type[Environment]], env_name: str) -> type[Environment]:
    env_cls = env_map.get(env_name.lower())
    if env_cls is None:
        raise HTTPException(404, f"unknown environment {env_name!r}")
    return env_cls

async def extract_sid(request: Request) -> str:
    x_session_id = request.headers.get("X-Session-ID")
    if not x_session_id:
        raise HTTPException(400, "X-Session-ID header is required")
    return x_session_id.strip()

class _LoggingRoute(APIRoute):
    def get_route_handler(self):
        original = super().get_route_handler()
        async def handler(request: Request):
            try:
                return await original(request)
            except Exception as exc:
                logging.exception(f"{request.method} {str(request.url)}", exc_info=exc)
                raise
        return handler

async def _delete_session(
    sid: str,
    active_envs: dict[str, Environment | None],
    last_ping: dict[str, float],
    setup_tasks: dict[str, asyncio.Task],
    ready: dict[str, asyncio.Event],
    setup_errors: dict[str, Exception],
):
    last_ping.pop(sid, None) # stop TTL

    task = setup_tasks.pop(sid, None) # cancel setup if it's running
    if task and not task.done():
        task.cancel()
        with suppress(asyncio.CancelledError):
            await task

    evt = ready.get(sid) # set error for waiters
    if evt and not evt.is_set():
        setup_errors[sid] = HTTPException(410, "Session deleted")
        evt.set()
    ready.pop(sid, None)

    env = active_envs.pop(sid, None) # teardown env
    if env:
        try:
            await maybe_await(env.teardown())
        except Exception:
            logging.exception("teardown failed for %s", sid)

    setup_errors.pop(sid, None) # clear error for waiters

class Server:
    def __init__(self, environments: list[type[Environment]]) -> None:
        """
        Environment hosting server. The first environment in the provided list is the default,
        and is served at /
        """
        if not environments:
            raise ValueError("Server requires at least one environment to be provided.")
        
        self._env_classes: dict[str, type[Environment]] = _get_env_map(environments)
        
        # Validate each environment
        for env_name, env_cls in self._env_classes.items():
            # Check that at least one tool is defined
            tools = env_cls.list_tools().tools
            if not tools:
                raise ValueError(f"Environment '{env_name}' has no tools defined. Add at least one @tool method.")
            
            # Check that at least one split is defined
            splits = env_cls.list_splits()
            if not splits:
                raise ValueError(f"Environment '{env_name}' has no splits defined. Implement list_splits() to return at least one split.")
            
            # Check that each split has at least one task
            # TODO: decide if we want to enforce this
            # for split in splits:
            #     tasks = env_cls.list_tasks(split)
            #     if not tasks:
            #         raise ValueError(f"Environment '{env_name}' has no tasks for split '{split}'. Ensure list_tasks() returns at least one task per split.")
        
        self._active_envs: dict[str, Environment | None] = {}
        self._last_ping: dict[str, float] = {}
        
        self._setup_tasks: dict[str, asyncio.Task] = {}
        self._ready: dict[str, asyncio.Event] = {}
        self._setup_errors: dict[str, Exception] = {}

        self._reaper_task: asyncio.Task | None = None

        async def await_environment_ready(sid: str) -> Environment:
            evt = self._ready.get(sid)
            if evt is None:
                raise HTTPException(404, "Active environment not found")

            if evt.is_set(): # fast path
                err = self._setup_errors.get(sid)
                if err:
                    raise err
                env = self._active_envs.get(sid)
                if env is None:
                    raise HTTPException(410, "Session deleted")
                return env

            await evt.wait()
            err = self._setup_errors.get(sid, None)
            if err:
                raise err
            env = self._active_envs.get(sid)
            if env is None:
                raise HTTPException(410, "Session deleted")
            return env

        async def reaper_coro(reaper_interval: float=5, stale_threshold: float=900):
            while True:
                await asyncio.sleep(reaper_interval)
                now = time.monotonic()
                stale = [sid for sid, ts in list(self._last_ping.items()) if now - ts > stale_threshold]
                for sid in stale:
                    await _delete_session(
                        sid,
                        self._active_envs,
                        self._last_ping,
                        self._setup_tasks,
                        self._ready,
                        self._setup_errors,
                    )

        @asynccontextmanager
        async def lifespan(app: FastAPI):
            self._reaper_task = asyncio.create_task(reaper_coro())
            yield
            if self._reaper_task:
                self._reaper_task.cancel()
                try:
                    await self._reaper_task
                except asyncio.CancelledError:
                    pass

        app = FastAPI(lifespan=lifespan)
        app.router.route_class = _LoggingRoute

        @app.exception_handler(Exception)
        async def unhandled_exception_handler(request: Request, exc: Exception):
            logging.exception(traceback.format_exc())
            raise HTTPException(500, "Internal server error")

        @app.get("/health")
        async def health():
            return {"status": "ok"}

        @app.post("/ping")
        async def ping(sid: str = Depends(extract_sid)):
            if sid not in self._active_envs:
                raise HTTPException(404, "Active environment not found")
            self._last_ping[sid] = time.monotonic()
            return {"status": "ok"}

        @app.get("/list_environments")
        async def get_envs():
            return list(self._env_classes.keys())

        @app.get("/{env_name}/tools")
        async def list_tools(env_name: str):
            env_cls = _get_env_cls(self._env_classes, env_name)
            return env_cls.list_tools().model_dump()

        @app.get("/{env_name}/splits")
        async def list_splits(env_name: str):
            env_cls = _get_env_cls(self._env_classes, env_name)
            return env_cls.list_splits()

        @app.post("/{env_name}/tasks")
        async def list_tasks(env_name: str, list_tasks: ListTasks):
            env_cls = _get_env_cls(self._env_classes, env_name)
            
            if list_tasks.split not in env_cls.list_splits():
                raise HTTPException(status_code=400, detail="Invalid split")

            return {"tasks": env_cls.list_tasks(list_tasks.split), "env_name": env_name}
        
        @app.post("/create")
        async def create_environment(create_session: CreateSession, sid: str = Depends(extract_sid)):
            if sid in self._active_envs:
                raise HTTPException(status_code=400, detail="Session already exists")

            self._active_envs[sid] = None
            self._last_ping[sid] = time.monotonic()
            self._ready[sid] = asyncio.Event()

            async def perform_setup():
                try:
                    env_cls = _get_env_cls(self._env_classes, create_session.env_name)
                    env = env_cls(task_spec=create_session.task_spec, secrets=create_session.secrets)
                    await maybe_await(env.setup())
                    self._active_envs[sid] = env
                except Exception as e:
                    self._setup_errors[sid] = e
                finally:
                    self._ready[sid].set()

            if sid not in self._setup_tasks:
                self._setup_tasks[sid] = asyncio.create_task(perform_setup())

            return {"sid": sid}

        async def require_existing_session(
            sid: str = Depends(extract_sid),
        ):
            self._last_ping[sid] = time.monotonic()
            return await await_environment_ready(sid)


        @app.post("/delete")
        async def delete_environment(sid: str = Depends(extract_sid)):
            if sid not in self._setup_tasks:
                raise HTTPException(404, "Active environment not found")
            await _delete_session(
                sid, 
                self._active_envs, 
                self._last_ping,
                self._setup_tasks,
                self._ready,
                self._setup_errors,
            )
            return {"sid": sid}

        @app.post("/create_session")
        async def create_session():
            return {"sid": str(uuid.uuid4())}
        @app.post("/delete_session")
        async def delete_session(sid: str = Depends(extract_sid)):
            return {"sid": sid}

        @app.post("/{env_name}/call")
        async def call_tool(request: Request, tool_call: ToolCall, env: Environment = Depends(require_existing_session)):
            async def call_tool_coro():
                res = await env._call_tool(tool_call.name, tool_call.input)
                return res.model_dump_json(indent=None)
            
            return EventSourceResponse(
                sse_task_stream(
                    lambda: call_tool_coro(),
                    request,
                    task_id=tool_call.task_id,
                ),
                ping=10,
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                }
            )

        @app.get("/{env_name}/prompt")
        async def get_prompt(request: Request, env: Environment = Depends(require_existing_session)) -> Blocks:
            return (await maybe_await(env.get_prompt()))

        # After all routes are defined, before adding the middleware:
        root_paths = set()
        for route in app.routes:
            path = getattr(route, 'path', '')
            segments = path.strip('/').split('/', 1)
            first_segment = segments[0] if segments else ''
            # Root-level paths don't start with a path parameter like {env_name}
            if not first_segment.startswith('{'):
                root_paths.add(first_segment.lower())

        @app.middleware("http")
        async def redirect_to_default_env(request: Request, call_next):
            segments = request.url.path.strip("/").split("/", 1)
            first_segment = segments[0].lower()
            
            if first_segment in root_paths or first_segment in self._env_classes:
                return await call_next(request)
            
            if not self._env_classes:
                return await call_next(request)
            
            first_env = next(iter(self._env_classes.keys()))
            new_path = f"/{first_env}{request.url.path}"
            if request.url.query:
                new_path += f"?{request.url.query}"
            
            return RedirectResponse(url=new_path, status_code=308)

        self.app = app

    def run(self, host: str = "0.0.0.0", port: int = 8080) -> None:
        uvicorn.run(self.app, host=host, port=port)


from dataclasses import dataclass
from typing import List, Literal, Optional, Sequence, Union

from .models import NormalizedEvent
from .utils import get


@dataclass
class UserMessage:
    content: str
    type: Literal['user_message'] = 'user_message'

@dataclass
class AssistantMessage:
    content: str
    type: Literal['assistant_message'] = 'assistant_message'

@dataclass
class SystemMessage:
    content: str
    type: Literal['system_message'] = 'system_message'

@dataclass
class ReasoningItem:
    content: Optional[str] = None
    content_reference: Optional[str] = None
    summary: Optional[str] = None
    type: Literal['reasoning'] = 'reasoning'

@dataclass
class ToolCall:
    name: str
    content: str
    call_id: str
    type: Literal['tool_call'] = 'tool_call'

@dataclass
class ToolResult:
    content: str
    call_id: str
    type: Literal['tool_result'] = 'tool_result'

UploadType = Union[UserMessage, AssistantMessage, SystemMessage, ReasoningItem, ToolCall, ToolResult]

def base_to_normalized(events: Sequence[UploadType]) -> List[NormalizedEvent]:
    normalized_events: List[NormalizedEvent] = []
    for event in events:
        normalized_events.append(NormalizedEvent(
            type=get(event, 'type'),
            content=get(event, 'content'),
            content_reference=get(event, 'content_reference'),
            summary=get(event, 'summary'),
            name=get(event, 'name'),
            call_id=get(event, 'call_id'),
        ))
    return normalized_events



import asyncio
import atexit
import multiprocessing as mp
import time
import uuid
from queue import Empty
from typing import Any, Dict, Optional

from anthropic.types import MessageParam as AnthropicMessageParam
from google.genai import types as gdm_types
from openai.types.responses import ResponseInputItemParam

from openreward.models import (
    InputEvent,
    LogMessageEvent,
    OutputEvent,
    RolloutConfig,
    RolloutStartedEvent,
    SendLoopConfig,
    ShutdownEvent,
)
from .background import worker_loop
from .serializers.ant import serialize_anthropic_message
from .serializers.base import UploadType, base_to_normalized
from .serializers.gdm import serialize_gdm_message
from .serializers.models import NormalizedEvent
from .serializers.oai_completions import (
    OpenAIChatMessage,
    openai_completions_to_normalized,
)
from .serializers.oai_responses import serialize_openai_response


class RolloutAPI:
    def __init__(self, 
        send_loop_config: SendLoopConfig,
        shutdown_timeout: float = 10.0,
        process_name: str = "openreward_rollout_api"
    ):
        self.send_loop_config = send_loop_config
        self.shutdown_timeout = shutdown_timeout
        self.process_name = process_name

        self._in: mp.Queue[InputEvent] = mp.Queue()
        self._out: mp.Queue[OutputEvent] = mp.Queue()

        self._p = mp.Process(
            target=worker_loop, 
            args=(self._in, self._out, self.send_loop_config),
            name=self.process_name,
            daemon=True
        )
        self._p.start()
        self.closed = asyncio.Event()

        atexit.register(self._atexit_handler)

    def create(
        self, 
        run_name: str, 
        rollout_name: Optional[str] = None, 
        environment: Optional[str] = None, 
        split: Optional[str] = None, 
        metadata: Optional[dict] = None, 
        task_spec: Optional[Dict[str, Any]] = None
    ) -> "Rollout":
        config = RolloutConfig(
            run_name=run_name,
            rollout_name=rollout_name,
            environment=environment,
            split=split,
            metadata=metadata,
            task_spec=task_spec
        )
        return Rollout(config=config, _in=self._in)

    def close(self):
        # check if we have the rollout api
        if self.closed.is_set():
            return
        self._in.put(ShutdownEvent())
        try:
            response = self._out.get(timeout=self.shutdown_timeout)
        except Empty:
            pass

        self._p.join(timeout=self.shutdown_timeout)
        if self._p.is_alive():
            self._p.terminate()
            print("Warning: terminating upload process while alive, some requests may not have been processed!")
        self.closed.set()

    def __del__(self):
        self.close()

    def _atexit_handler(self):
        # Best effort if user forgot
        try:
            self.close()
        except Exception:
            pass

class Rollout:
    def __init__(self, config: RolloutConfig, _in: mp.Queue):
        self.config = config
        self.event_id = str(uuid.uuid4())
        self._in = _in
        self._in.put(RolloutStartedEvent.from_config(
            event_id=self.event_id,
            timestamp=int(time.time() * 1000),
            config=config
        ))
        self.logged_messages = 0

    def _log_message(self, normalized_event: NormalizedEvent, reward: Optional[float] = None, is_finished: Optional[bool] = False, metadata: Optional[dict] = None):
        
        if is_finished is None:
            is_finished = False
        
        self._in.put(LogMessageEvent(
            eventId = str(uuid.uuid4()),
            timestamp=int(time.time() * 1000),
            index=self.logged_messages,
            rolloutEventId=self.event_id,
            type=normalized_event.type,
            content=normalized_event.content,
            contentReference=normalized_event.content_reference,
            summary=normalized_event.summary,
            name=normalized_event.name,
            callId=normalized_event.call_id,
            reward=reward,
            is_finished=is_finished,
            metadata=metadata or {},
        ))
        self.logged_messages += 1

    def log(
        self, 
        message: UploadType,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        normalized_event = base_to_normalized([message])[0]
        self._log_message(normalized_event, reward, is_finished, metadata)

    def log_openai_completions(
        self,
        message: OpenAIChatMessage,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        normalized_event = openai_completions_to_normalized([message])[0]
        self._log_message(normalized_event, reward, is_finished, metadata)

    def log_openai_response(
        self,
        message: ResponseInputItemParam,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        event = serialize_openai_response(message)
        self._log_message(event, reward, is_finished, metadata)

    def log_anthropic_message(
        self,
        message: AnthropicMessageParam,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        for event in serialize_anthropic_message(message):
            self._log_message(event, reward, is_finished, metadata)

    def log_gdm_message(
        self,
        message: gdm_types.Content,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        for event in serialize_gdm_message(message):
            self._log_message(event, reward, is_finished, metadata)



"""Command line utilities for managing OpenReward environments."""
from __future__ import annotations

import argparse
import re
from dataclasses import dataclass, field
from importlib import resources
from pathlib import Path
from string import Template
from typing import Callable, Mapping


@dataclass(frozen=True)
class TemplateFiles:
    """Collection of scaffold files for an environment."""

    dockerfile: str
    server_py: str
    extra_files: Mapping[str, str] = field(default_factory=dict)


def _pascal_case(name: str) -> str:
    parts = re.split(r"[^0-9a-zA-Z]+", name)
    cleaned = "".join(part.capitalize() for part in parts if part)
    if not cleaned:
        cleaned = "Environment"
    if cleaned[0].isdigit():
        cleaned = f"Env{cleaned}"
    return cleaned


def _load_template_file(template_name: str, filename: str) -> str:
    package = f"openreward.templates.{template_name}"
    file = resources.files(package).joinpath(filename)
    return file.read_text(encoding="utf-8")


def _basic_template(env_name: str) -> TemplateFiles:
    class_name = _pascal_case(env_name)
    dockerfile = _load_template_file("basic", "Dockerfile")
    server_template = Template(_load_template_file("basic", "server.py.tmpl"))
    server_py = server_template.substitute(CLASS_NAME=class_name)
    extra_files = {
        "requirements.txt": _load_template_file("basic", "requirements.txt"),
    }
    return TemplateFiles(dockerfile=dockerfile, server_py=server_py, extra_files=extra_files)


TemplateFactory = Callable[[str], TemplateFiles]


TEMPLATES: Mapping[str, TemplateFactory] = {
    "basic": _basic_template,
}


def _render_template(template_name: str, env_name: str) -> TemplateFiles:
    generator = TEMPLATES.get(template_name)
    if generator is None:
        raise ValueError(f"Unknown template '{template_name}'. Available templates: {', '.join(TEMPLATES)}")
    return generator(env_name)


def _write_file(path: Path, contents: str) -> None:
    if path.exists():
        raise FileExistsError(f"{path} already exists")
    path.write_text(contents, encoding="utf-8")


def command_init(environment: str, template: str) -> None:
    target_dir = Path.cwd() / environment
    target_dir.mkdir(parents=True, exist_ok=False)

    files = _render_template(template, environment)

    _write_file(target_dir / "Dockerfile", files.dockerfile)
    _write_file(target_dir / "server.py", files.server_py)
    for relative, contents in files.extra_files.items():
        _write_file(target_dir / relative, contents)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="orwd", description="OpenReward CLI utilities")
    subparsers = parser.add_subparsers(dest="command", required=True)

    init_parser = subparsers.add_parser(
        "init",
        help="Initialize a new environment template",
    )
    init_parser.add_argument("environment", help="Name of the environment to scaffold")
    init_parser.add_argument(
        "--template",
        default="basic",
        choices=sorted(TEMPLATES.keys()),
        help="Scaffold template to use (default: %(default)s)",
    )
    init_parser.set_defaults(func=lambda args: command_init(args.environment, args.template))

    return parser


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    try:
        args.func(args)
    except FileExistsError as exc:
        parser.error(str(exc))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import inspect
from typing import Awaitable, TypeVar

T = TypeVar("T")

async def maybe_await(x: T | Awaitable[T]) -> T:
    if inspect.isawaitable(x):
        return await x
    else:
        return x